#define BUILD_NUMBER \
	"build-1031360"
#define BUILD_NUMBER_NUMERIC \
	1031360
#define BUILD_NUMBER_NUMERIC_STRING \
	"1031360"
#define PRODUCT_BUILD_NUMBER \
	"product-build-21136"
#define PRODUCT_BUILD_NUMBER_NUMERIC \
	21136
#define PRODUCT_BUILD_NUMBER_NUMERIC_STRING \
	"21136"
